"""CLI for League Manager tool."""

import json
from typing import Annotated, Optional

import typer
from advanced_alchemy.exceptions import DuplicateKeyError
from rich import print
from rich.prompt import Prompt
from sqlalchemy import delete
from sqlalchemy.orm import Session

from leaguemanager import models
from leaguemanager.core import get_settings
from leaguemanager.db import register_sqlite
from leaguemanager.db.cli.alembic_cli import db_app
from leaguemanager.dependency.cli_callbacks import (
    provide_manager_service,
    provide_sync_db_session,
)
from leaguemanager.services import (
    FixtureService,
    LeagueService,
    OrganizationService,
    SeasonService,
    TeamMembershipService,
    TeamService,
)

from . import __app_name__, __version__

settings = get_settings()
register_sqlite()

app = typer.Typer()


def _version_callback(value: bool) -> None:
    if value:
        print(f"{__app_name__} v{__version__}")
        raise typer.Exit()


@app.command(help="Populate the database with synthetic data.")
def populate(
    season_service: Annotated[
        Optional[SeasonService], typer.Argument(callback=provide_manager_service, parser=SeasonService)
    ] = None,
    league_service: Annotated[
        Optional[LeagueService], typer.Argument(callback=provide_manager_service, parser=LeagueService)
    ] = None,
    team_service: Annotated[
        Optional[TeamService], typer.Argument(callback=provide_manager_service, parser=TeamService)
    ] = None,
    team_membership_service: Annotated[
        Optional[TeamMembershipService],
        typer.Argument(callback=provide_manager_service, parser=TeamMembershipService),
    ] = None,
    organization_service: Annotated[
        Optional[OrganizationService],
        typer.Argument(callback=provide_manager_service, parser=OrganizationService),
    ] = None,
) -> None:
    if (settings.APP_DIR / "example_data.json").exists():
        with open(settings.APP_DIR / "example_data.json") as _data:
            data = json.load(_data)
    elif not (settings.SYNTH_DATA_DIR / "example_data.json").exists():
        data_file = Prompt.ask("Please provide the path to the data directory: ")
        try:
            with open(data_file) as _data:
                data = json.load(_data)
        except FileNotFoundError:
            print(f"File {data_file} does not exist.")
            return
    else:
        with open(settings.SYNTH_DATA_DIR / "example_data.json") as _data:
            data = json.load(_data)
    try:
        org = organization_service.create(data["organization"], auto_commit=True)
        for league in data["leagues"]:
            league["organization_id"] = org.id
            league_service.create(league, auto_commit=True)
        for season in data["seasons"]:
            _league = league_service.get_one(name=season["league"])
            season["league_id"] = _league.id
            season_service.create(
                season,
                auto_commit=True,
            )
        for team in data["teams"]:
            _team_membership = team_membership_service.create(
                {"label": f"{team['name']} - {team['season']}"},
                auto_commit=True,
            )
            team["team_membership_id"] = _team_membership.id
            team.pop("season", None)  # Remove season key if it exists

        for team in data["teams"]:
            team_service.create(
                team,
                auto_commit=True,
            )
    except DuplicateKeyError:
        print("[red]Data already exists in the database. Will not create any data...[/red]")
        return

    print(
        "✨ [green]Successfully created data![/green] ✨",
        f"\nCreated {season_service.count()} Seasons",
        f"\nCreated {league_service.count()} Leagues",
        f"\nCreated {team_service.count()} Teams",
        f"\nCreated {organization_service.count()} Organization",
    )
    return


@app.command(help="Check the counts of data in each table.")
def check(
    session: Annotated[Optional[Session], typer.Argument(callback=provide_sync_db_session, parser=Session)] = None,
    season_service: Annotated[
        Optional[SeasonService], typer.Argument(callback=provide_manager_service, parser=SeasonService)
    ] = None,
    league_service: Annotated[
        Optional[LeagueService], typer.Argument(callback=provide_manager_service, parser=LeagueService)
    ] = None,
    team_service: Annotated[
        Optional[TeamService], typer.Argument(callback=provide_manager_service, parser=TeamService)
    ] = None,
    organization_service: Annotated[
        Optional[OrganizationService],
        typer.Argument(callback=provide_manager_service, parser=OrganizationService),
    ] = None,
    all: bool = typer.Option(
        False,
        "--all",
        "-a",
        help="Check all seasons, leagues, and teams.",
    ),
) -> None:
    if all:
        print(
            "✨ [green]These are the counts of data in each table[/green] ✨",
            f"\n{season_service.count()} Seasons",
            f"\n{league_service.count()} Leagues",
            f"\n{team_service.count()} Teams",
            f"\n{organization_service.count()} Organization",
        )
    else:
        model = Prompt.ask("Please provide the model name to check (i.e. Season): ")
        try:
            _models = session.query(getattr(models, model)).all()
            print(f"Found {session.query(getattr(models, model)).count()} {model}s.")
            for m in _models:
                if m.name:
                    print(f">> {m.name}")
        except AttributeError:
            print(f"Model [green]{model}[/green] does not exist.")
    return


@app.command(name="delete", help="Delete data from the database.")
def _delete(
    session: Annotated[Optional[Session], typer.Argument(callback=provide_sync_db_session, parser=Session)] = None,
    all: bool = typer.Option(
        False,
        "--all",
        "-a",
        help="Delete all seasons, leagues, teams, and organization.",
    ),
) -> None:
    if all:
        session.execute(delete(models.Season))
        session.execute(delete(models.League))
        session.execute(delete(models.Team))
        session.execute(delete(models.Organization))
        session.commit()
        print("⛔ [yellow]Removed all Seasons, Leagues, Teams, and Organizations.[/yellow]⛔")
    else:
        model = Prompt.ask("Please provide the model name to delete (i.e. Season): ")
        try:
            session.execute(delete(getattr(models, model.title())))
            session.commit()
        except AttributeError:
            print(f"Model [green]{model.title()}[/green] does not exist.")
        print(f"⛔ [yellow]Removed all data in the {model.title()} table.[/yellow]⛔")
    return


@app.callback(no_args_is_help=True, help="League Manager CLI.")
def main(
    ctx: typer.Context,
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            "-V",
            callback=_version_callback,
            is_eager=True,
            help="Show the version and exit.",
        ),
    ] = None,
):
    """League Manager CLI."""
    return


# This creates an app for Typer, adding the above commands.
# This enables us to add the `db_app`, which is a Click group from Advanced Alchemy.
app = typer.main.get_command(app)
app.add_command(db_app)



from .fixture import Fixture
from .league import League
from .league_properties import LeagueProperties
from .organization import Organization
from .phase import Phase
from .ruleset import Ruleset
from .season import Season
from .site import Site

__all__ = [
    "Fixture",
    "League",
    "LeagueProperties",
    "Organization",
    "Phase",
    "Season",
    "Ruleset",
    "Site",
]



"""Sqlite3 datetime adapters; import this module to register adapters for datetime objects;
these were built in before Python 3.12 but are deprecated in 3.12"""

import datetime
import sqlite3

# Reference: https://docs.python.org/3/library/sqlite3.html?highlight=sqlite3#sqlite3-adapter-converter-recipes


def adapt_date_iso(val):
    """Adapt datetime.date to ISO 8601 date."""
    return val.isoformat()


def adapt_datetime_iso(val):
    """Adapt datetime.datetime to timezone-naive ISO 8601 date."""
    return val.isoformat()


def adapt_datetime_epoch(val):
    """Adapt datetime.datetime to Unix timestamp."""
    return int(val.timestamp())


def convert_date(val):
    """Convert ISO 8601 date to datetime.date object."""
    return datetime.date.fromisoformat(val.decode())


def convert_datetime(val):
    """Convert ISO 8601 datetime to datetime.datetime object."""
    return datetime.datetime.fromisoformat(val.decode())


def convert_timestamp(val):
    """Convert Unix epoch timestamp to datetime.datetime object."""
    return datetime.datetime.fromtimestamp(int(val))


def register_adapters():
    """Register adapters for datetime objects."""
    sqlite3.register_adapter(datetime.date, adapt_date_iso)
    sqlite3.register_adapter(datetime.datetime, adapt_datetime_iso)
    sqlite3.register_adapter(datetime.datetime, adapt_datetime_epoch)


def register_converters():
    """Register converters for datetime objects."""
    sqlite3.register_converter("date", convert_date)
    sqlite3.register_converter("datetime", convert_datetime)
    sqlite3.register_converter("timestamp", convert_timestamp)


def register_sqlite():
    """Register adapters and converters for datetime objects."""
    register_adapters()
    register_converters()



from advanced_alchemy.repository import (
    SQLAlchemySyncRepository,
    SQLAlchemySyncSlugRepository,
)

from leaguemanager import models

__all__ = [
    "AddressSyncRepository",
    "CitySyncRepository",
    "CountrySyncRepository",
    "StateSyncRepository",
    "RoleSyncRepository",
    "UserSyncRepository",
    "UserRoleSyncRepository",
    "AthleteSyncRepository",
    "AthleteStatsSyncRepository",
    "FixtureSyncRepository",
    "IndividualMembershipSyncRepository",
    "LeagueSyncRepository",
    "LeaguePropertiesSyncRepository",
    "ManagerSyncRepository",
    "ManagerMembershipSyncRepository",
    "ManagingSyncRepository",
    "OfficialSyncRepository",
    "OfficiatingSyncRepository",
    "OrganizationSyncRepository",
    "PhaseSyncRepository",
    "RulesetSyncRepository",
    "SeasonSyncRepository",
    "SiteSyncRepository",
    "TeamSyncRepository",
    "TeamMembershipSyncRepository",
    "TeamStatsSyncRepository",
]


class AddressSyncRepository(SQLAlchemySyncRepository[models.Address]):
    """Address repository."""

    model_type = models.Address


class CitySyncRepository(SQLAlchemySyncRepository[models.City]):
    """City repository."""

    model_type = models.City


class CountrySyncRepository(SQLAlchemySyncRepository[models.Country]):
    """Country repository."""

    model_type = models.Country


class StateSyncRepository(SQLAlchemySyncRepository[models.State]):
    """State repository."""

    model_type = models.Country


class RoleSyncRepository(SQLAlchemySyncSlugRepository[models.Role]):
    """Role repository."""

    model_type = models.Role


class UserSyncRepository(SQLAlchemySyncRepository[models.User]):
    """User repository."""

    model_type = models.User


class UserRoleSyncRepository(SQLAlchemySyncRepository[models.UserRole]):
    """UserRole repository."""

    model_type = models.UserRole


class AthleteSyncRepository(SQLAlchemySyncRepository[models.Athlete]):
    """Athlete repository."""

    model_type = models.Athlete


class AthleteStatsSyncRepository(SQLAlchemySyncRepository[models.AthleteStats]):
    """AthleteStats repository."""

    model_type = models.AthleteStats


class FixtureSyncRepository(SQLAlchemySyncRepository[models.Fixture]):
    """Fixture repository."""

    model_type = models.Fixture


class IndividualMembershipSyncRepository(SQLAlchemySyncRepository[models.IndividualMembership]):
    """IndividualMembership repository."""

    model_type = models.IndividualMembership


class LeagueSyncRepository(SQLAlchemySyncSlugRepository[models.League]):
    """League repository."""

    model_type = models.League


class LeaguePropertiesSyncRepository(SQLAlchemySyncSlugRepository[models.LeagueProperties]):
    """LeagueProperties repository."""

    model_type = models.LeagueProperties


class ManagerSyncRepository(SQLAlchemySyncRepository[models.Manager]):
    """Manager repository."""

    model_type = models.Manager


class ManagerMembershipSyncRepository(SQLAlchemySyncRepository[models.ManagerMembership]):
    """ManagerMembership repository."""

    model_type = models.ManagerMembership


class ManagingSyncRepository(SQLAlchemySyncRepository[models.Managing]):
    """Managing repository."""

    model_type = models.Managing


class OfficialSyncRepository(SQLAlchemySyncRepository[models.Official]):
    """Official repository."""

    model_type = models.Official


class OfficiatingSyncRepository(SQLAlchemySyncRepository[models.Officiating]):
    """Officiating repository."""

    model_type = models.Officiating


class OrganizationSyncRepository(SQLAlchemySyncSlugRepository[models.Organization]):
    """Organization repository."""

    model_type = models.Organization


class PhaseSyncRepository(SQLAlchemySyncRepository[models.Phase]):
    """Phase repository."""

    model_type = models.Phase


class RulesetSyncRepository(SQLAlchemySyncRepository[models.Ruleset]):
    """Ruleset repository."""

    model_type = models.Ruleset


class SeasonSyncRepository(SQLAlchemySyncSlugRepository[models.Season]):
    """Season repository."""

    model_type = models.Season


class SiteSyncRepository(SQLAlchemySyncSlugRepository[models.Site]):
    """Site repository."""

    model_type = models.Site


class TeamSyncRepository(SQLAlchemySyncSlugRepository[models.Team]):
    """Team repository."""

    model_type = models.Team


class TeamMembershipSyncRepository(SQLAlchemySyncSlugRepository[models.TeamMembership]):
    """TeamMembership repository."""

    model_type = models.TeamMembership


class TeamStatsSyncRepository(SQLAlchemySyncSlugRepository[models.TeamStats]):
    """TeamStats repository."""

    model_type = models.TeamStats



from __future__ import annotations

import typer
from sqlalchemy.orm import Session

from leaguemanager.dependency import LeagueManager
from leaguemanager.services._typing import SQLAlchemyAsyncConfigT, SQLAlchemySyncConfigT, SyncRepositoryT, SyncServiceT

registry = LeagueManager()


def provide_manager_service(param: typer.CallbackParam) -> SyncServiceT | SyncRepositoryT:
    return registry.provide_db_service(service_type=param.type.func)


def provide_manager_repository(param: typer.CallbackParam) -> SyncRepositoryT:
    return registry.provide_db_repository(repository_type=param.type.func)


def provide_sync_db_session() -> Session:
    return registry.provide_db_session


def provide_sync_db_config() -> SQLAlchemySyncConfigT:
    return registry.provide_sync_config


def provide_async_db_config() -> SQLAlchemyAsyncConfigT:
    return registry.provide_async_config



from __future__ import annotations

from datetime import UTC, datetime

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, String, Table
from sqlalchemy import Enum as SA_Enum
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import PropertiesBase, mapper, metadata
from leaguemanager.models.enums import Category, MatchDay

if TYPE_CHECKING:
    from uuid import UUID

    from .organization import Organization


def category_converter(value):
    if isinstance(value, Category):
        return value.name
    return value


@define(slots=False)
class LeagueProperties(PropertiesBase):
    """Defines the properties specific to a Competition, such as a League.

    It broadly describes qualifiers for a competition, which can include the sport, age class,
    day of competition, or other categories (such as "Coed" or "Division 2").

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        site_id (UUID): ForeignKey to Site.
        league_id (UUID): ForeignKey to League.
        sport (str): The sport type of the League.
        category (str): Defines if league is Women, Men, or Coed.
        division (str): Indicates the "Division" of a League (i.e., "Div 3", "B League", etc...)
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    site_id: UUID | None = field(default=None)
    league_id: UUID | None = field(default=None)
    sport: str | None = field(default=None, validator=validators.max_len(16))
    category: str | None = field(
        default=None,
        converter=category_converter,
        validator=validators.optional(validators.in_({a.name for a in Category})),
    )
    division: str | None = field(default=None, validator=validators.max_len(12))


# SQLAlchemy Imperative Mappings

league_properties = Table(
    "league_properties",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("site_id", SA_UUID, ForeignKey("site.id"), nullable=True),
    Column("league_id", SA_UUID, ForeignKey("league.id"), nullable=True),
    Column("sport", String(16), nullable=True),
    Column("category", SA_Enum(Category), nullable=True),
    Column("division", String(12), nullable=True),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    LeagueProperties,
    league_properties,
    properties={"league": relationship("League", back_populates="league_properties")},
)



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import MembershipBase, mapper, metadata


@define(slots=False)
class ManagerMembership(MembershipBase):
    """An manager's affiliation with a Team.

    Instead of linking an Manager object directly to a team, the Manager is linked
    to an ManagerMembership object, which defines their specific affiliation to a
    given team. That way, an Manager can be linked to different ManagerMembership
    objects, each connected to different Teams (in the event that an Manager has multiple
    team affiliations). The ManagerMembership object can define properties specific
    to a Team, such as management role, start and end dates, etc...

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        team_id (UUID): ForeignKey to Team.
        manager_id (UUID): ForeignKey to Manager.
        label (str): Label describing the Manager association (i.e., Henry Cooper, Team A, etc...)
        role (str): Role within the team (i.e., Head Coach, Assistant, Trainer, etc...)
        start_date (datetime): Date that individual joined Team.
        end_date (datetime): Date that individual left Team.
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    team_id: UUID | None = field(default=None)
    manager_id: UUID | None = field(default=None)
    label: str | None = field(default=None, validator=validators.optional(validators.max_len(24)))
    role: str | None = field(default=None, validator=validators.optional(validators.max_len(12)))
    start_date: datetime | None = field(default=None)
    end_date: datetime | None = field(default=None)


# SQLAlchemy Imperative Mapping

manager_membership = Table(
    "manager_membership",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("team_id", SA_UUID, ForeignKey("team.id"), nullable=True),
    Column("manager_id", SA_UUID, ForeignKey("manager.id"), nullable=True),
    Column("label", String(24), nullable=True),
    Column("role", String(12), nullable=True),
    Column("start_date", DateTime(), nullable=True, default=None),
    Column("end_date", DateTime(), nullable=True, default=None),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    ManagerMembership,
    manager_membership,
    properties={
        "manager": relationship("Manager", back_populates="manager_membership", uselist=False),
        "team": relationship("Team", back_populates="manager_memberships"),
    },
)



from enum import Enum


class States(str, Enum):
    """States in the United States."""

    AL = "Alabama"
    AK = "Alaska"
    AZ = "Arizona"
    AR = "Arkansas"
    CA = "California"
    CO = "Colorado"
    CT = "Connecticut"
    DE = "Delaware"
    FL = "Florida"
    GA = "Georgia"
    HI = "Hawaii"
    ID = "Idaho"
    IL = "Illinois"
    IN = "Indiana"
    IA = "Iowa"
    KS = "Kansas"
    KY = "Kentucky"
    LA = "Louisiana"
    ME = "Maine"
    MD = "Maryland"
    MA = "Massachusetts"
    MI = "Michigan"
    MN = "Minnesota"
    MS = "Mississippi"
    MO = "Missouri"
    MT = "Montana"
    NE = "Nebraska"
    NV = "Nevada"
    NH = "New Hampshire"
    NJ = "New Jersey"
    NM = "New Mexico"
    NY = "New York"
    NC = "North Carolina"
    ND = "North Dakota"
    OH = "Ohio"
    OK = "Oklahoma"
    OR = "Oregon"
    PA = "Pennsylvania"
    RI = "Rhode Island"
    SC = "South Carolina"
    SD = "South Dakota"
    TN = "Tennessee"
    TX = "Texas"
    UT = "Utah"
    VT = "Vermont"
    VA = "Virginia"
    WA = "Washington"
    WV = "West Virginia"
    WI = "Wisconsin"
    WY = "Wyoming"



"""Base classes for database objects are loosely based on IPTC Sport Schema. see: https://sportschema.org"""

from __future__ import annotations

from datetime import UTC, datetime
from typing import TYPE_CHECKING, Any, Protocol
from uuid import UUID, uuid4

from attrs import converters, define, field, validators
from sqlalchemy import MetaData
from sqlalchemy.orm import Mapper, registry

from .enums import Gender

mapper = registry()
metadata = MetaData()

##############
# BASE OBJECTS
##############


@define(slots=False)
class UUIDBase:
    id: UUID = field(factory=uuid4)

    if TYPE_CHECKING:
        __name__: str
        __mapper__: Mapper[Any]

    def to_dict(self, exclude: set[str] | None = None) -> dict[str, Any]:
        """Convert model to dictionary.

        Returns:
            Dict[str, Any]: A dict representation of the model
        """
        exclude = {"sa_orm_sentinel", "_sentinel"}.union(self._sa_instance_state.unloaded).union(exclude or [])  # type: ignore[attr-defined]
        return {
            field: getattr(self, field)
            for field in self.__mapper__.columns.keys()  # noqa: SIM118
            if field not in exclude
        }

    def str_to_iso(date_string: str, format: str):
        """Converts string to datetime object."""
        return datetime.strptime(date_string, format)


@define(slots=False)
class UUIDAuditBase(UUIDBase):
    created_at: datetime = field(factory=lambda: datetime.now(UTC))
    updated_at: datetime = field(factory=lambda: datetime.now(UTC))


##########################
# COMPETITION OBJECT TYPES
##########################


@define(slots=False)
class GoverningBodyBase(UUIDAuditBase):
    """Base class for organization responsible for running competitions."""

    org_type: str | None = field(default="amateur-league", validator=validators.max_len(20))


@define(slots=False)
class CompetitionBase(UUIDAuditBase):
    """Base class for competitions. Can be child of Governing Body or other Competitions.

    The competition_type attribute should describe whether the object is meant to be a recurring
    construct (i.e., recurring-competition), or something finite (i.e., competition, tournament).
    The value of this property will vary based on the object that inherits this class.

    #TODO: May need to create a constraint with pre-determined types.
    """

    competition_type: str | None = field(default="recurring-competition", validator=validators.max_len(30))


@define(slots=False)
class CompetitionPhaseBase(CompetitionBase):
    """A sub-group of events within a competition (i.e., regular, semifinals, knockout round, etc...)"""

    phase_type: str | None = field(default="regular", validator=validators.max_len(30))


@define(slots=False)
class EventBase(UUIDAuditBase):
    """Base class for an event (match/fixture) that produces results."""

    ...


@define(slots=False)
class ActionBase(UUIDAuditBase):
    """Base class for actions related to an event."""

    ...


@define(slots=False)
class PropertiesBase(UUIDAuditBase):
    """Base class for rules or properties."""


#########################
# MEMBERSHIP OBJECT TYPES
#########################


@define(slots=False)
class AgentBase(UUIDAuditBase):
    """Base class for Individual, Team, or Club."""

    ...


@define(slots=False)
class IndividualBase(AgentBase):
    """All individuals will inherit these properties, but properties optional to allow anonymous usage."""

    first_name: str | None = field(default=None, validator=validators.optional(validators.max_len(20)))
    last_name: str | None = field(default=None, validator=validators.optional(validators.max_len(20)))
    middle_name: str | None = field(default=None, validator=validators.optional(validators.max_len(20)))
    full_name: str | None = field(default=None, validator=validators.optional(validators.max_len(64)))
    alias: str | list[str] | None = field(default=None)
    email: str | None = field(default=None, validator=validators.optional(validators.max_len(40)))
    mobile_phone: str | None = field(default=None, validator=validators.optional(validators.max_len(16)))
    birth_date: datetime | None = field(default=None)
    birth_year: int | None = field(default=None)
    sex: str | None = field(default=None, validator=validators.optional(validators.in_(("male", "female"))))
    gender: str | None = field(default=None, validator=validators.optional(validators.in_((e.value for e in Gender))))


@define(slots=False)
class MembershipBase(UUIDAuditBase):
    """Defines membership of an entity in another entity for a given time."""

    ...


############################
# PARTICIPATION OBJECT TYPES
############################


@define(slots=None)
class SiteBase(UUIDBase):
    """Base class for a given site."""

    ...


@define(slots=None)
class ParticipationBase(UUIDAuditBase):
    """Base class for modeling participation of an actor in an event."""

    ...


@define(slots=None)
class TeamParticipationBase(ParticipationBase):
    """Base class for team stats in an Event."""

    ...


@define(slots=False)
class OfficialParticipationBase(ParticipationBase):
    """Base class for involvement of Officials in an Event."""

    ...


@define(slots=False)
class IndividualParticipationBase(ParticipationBase):
    """Base class for individual stats in an Event."""

    ...



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Ruleset
from leaguemanager.repository import RulesetSyncRepository
from leaguemanager.repository._async import RulesetAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["RulesetService", "RulesetAsyncService"]


class RulesetService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = RulesetSyncRepository


class RulesetAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = RulesetAsyncRepository



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Manager
from leaguemanager.repository import ManagerSyncRepository
from leaguemanager.repository._async import ManagerAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["ManagerService", "ManagerAsyncService"]


class ManagerService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = ManagerSyncRepository


class ManagerAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = ManagerAsyncRepository



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import LeagueProperties
from leaguemanager.repository import LeaguePropertiesSyncRepository
from leaguemanager.repository._async import LeaguePropertiesAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["LeaguePropertiesService", "LeaguePropertiesAsyncService"]


class LeaguePropertiesService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = LeaguePropertiesSyncRepository


class LeaguePropertiesAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = LeaguePropertiesAsyncRepository



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import TeamStats
from leaguemanager.repository import TeamStatsSyncRepository
from leaguemanager.repository._async import TeamStatsAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["TeamStatsService", "TeamStatsAsyncService"]


class TeamStatsService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = TeamStatsSyncRepository


class TeamStatsAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = TeamStatsAsyncRepository



import json
from typing import Any

import pytest

from leaguemanager.core import get_settings

settings = get_settings()


@pytest.fixture(name="all_data", autouse=True, scope="session")
def fx_all_data() -> dict[str, Any]:
    with open(settings.SYNTH_DATA_DIR / "example_data.json") as _data:
        data = json.load(_data)
    return data


@pytest.fixture(name="all_teams", autouse=True, scope="session")
def fx_all_teams(all_data) -> dict[str, Any]:
    return all_data["teams"]


@pytest.fixture(name="all_leagues", autouse=True, scope="session")
def fx_all_leagues(all_data) -> dict[str, Any]:
    return all_data["leagues"]


@pytest.fixture(name="all_seasons", autouse=True, scope="session")
def fx_all_seasons(all_data) -> dict[str, Any]:
    return all_data["seasons"]



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Site
from leaguemanager.repository import SiteSyncRepository
from leaguemanager.repository._async import SiteAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["SiteService", "SiteAsyncService"]


class SiteService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = SiteSyncRepository


class SiteAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = SiteAsyncRepository



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Organization
from leaguemanager.repository import OrganizationSyncRepository
from leaguemanager.repository._async import OrganizationAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["OrganizationService", "OrganizationAsyncService"]


class OrganizationService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = OrganizationSyncRepository


class OrganizationAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = OrganizationAsyncRepository



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from advanced_alchemy.config import SQLAlchemyAsyncConfig, SQLAlchemySyncConfig
from advanced_alchemy.repository import SQLAlchemyAsyncRepository, SQLAlchemySyncRepository
from advanced_alchemy.service import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

from leaguemanager.models.base import UUIDBase

ModelT = TypeVar("ModelT", bound=UUIDBase)
SyncRepositoryT = TypeVar("RepositoryT", bound=SQLAlchemySyncRepository)
AsyncRepositoryT = TypeVar("RepositoryT", bound=SQLAlchemyAsyncRepository)
SyncServiceT = TypeVar("ServiceT", bound=SQLAlchemySyncRepositoryService[ModelT])
AsyncServiceT = TypeVar("ServiceT", bound=SQLAlchemyAsyncRepositoryService[ModelT])
SQLAlchemySyncConfigT = TypeVar("ConfigT", bound=SQLAlchemySyncConfig)
SQLAlchemyAsyncConfigT = TypeVar("ConfigT", bound=SQLAlchemyAsyncConfig)



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import ParticipationBase, mapper, metadata

if TYPE_CHECKING:
    from leaguemanager.models.baseline import Team


@define(slots=False)
class Officiating(ParticipationBase):
    """The participation of an official in a given Fixture.

    Each Official that is participating in an Fixture has an Officiating object that
    represents their involvement in that specific Fixture. It can also be used to track
    attributes specific to the Fixture (i.e., payment, uniform color, etc...)

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        fixture_id (UUID): ForeignKey to Fixture.
        label (str): Description of Officiating object (i.e., "Ref Jacobs, Week 2, Amateur Mudwrestling", etc...)
        paid (bool): Whether official has been paid for the Fixture.
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    fixture_id: UUID | None = field(default=None)
    label: str | None = field(default=None, validator=validators.max_len(90))
    paid: bool | None = field(default=False)


# SQLAlchemy Imperative Mappings

officiating = Table(
    "officiating",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("fixture_id", SA_UUID, ForeignKey("fixture.id"), nullable=True),
    Column("label", String(90), nullable=False),
    Column("paid", Boolean),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    Officiating,
    officiating,
    properties={
        "fixture": relationship("Fixture", back_populates="officiating"),
        "officials": relationship("Official", back_populates="officiating"),
    },
)



from .fixture import FixtureAsyncService, FixtureService
from .league import LeagueAsyncService, LeagueService
from .league_properties import LeaguePropertiesAsyncService, LeaguePropertiesService
from .organization import OrganizationAsyncService, OrganizationService
from .phase import PhaseAsyncService, PhaseService
from .ruleset import RulesetAsyncService, RulesetService
from .season import SeasonAsyncService, SeasonService
from .site import SiteAsyncService, SiteService

__all__ = [
    "FixtureAsyncService",
    "FixtureService",
    "LeagueAsyncService",
    "LeagueService",
    "LeaguePropertiesAsyncService",
    "LeaguePropertiesService",
    "OrganizationAsyncService",
    "OrganizationService",
    "PhaseAsyncService",
    "PhaseService",
    "RulesetAsyncService",
    "RulesetService",
    "SeasonAsyncService",
    "SeasonService",
    "SiteService",
    "SiteAsyncService",
]



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import ParticipationBase, mapper, metadata

if TYPE_CHECKING:
    from leaguemanager.models.baseline import Team


@define(slots=False)
class Managing(ParticipationBase):
    """The participation of a Manager in a given Fixture.

    Each Manager that is participating in an Fixture has a Managing object that
    represents their involvement in that specific Fixture. It can also be used to track
    attributes specific to the Fixture.

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        fixture_id (UUID): ForeignKey to Fixture.
        label (str): Description of Managing object (i.e., "Ref Jacobs, Week 2, Amateur Mudwrestling", etc...)
        notes (str): Can be used to describe game specific notes.
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    fixture_id: UUID | None = field(default=None)
    label: str | None = field(default=None, validator=validators.max_len(40))
    notes: str | None = field(default=None, validator=validators.optional(validators.max_len(90)))


# SQLAlchemy Imperative Mappings

managing = Table(
    "managing",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("fixture_id", SA_UUID, ForeignKey("fixture.id"), nullable=True),
    Column("label", String(40), nullable=False),
    Column("notes", String(90), nullable=False),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    Managing,
    managing,
    properties={
        "fixture": relationship("Fixture", back_populates="managing"),
        "managers": relationship("Manager", back_populates="managing"),
    },
)



from .athlete import AthleteAsyncService, AthleteService
from .individual_membership import IndividualMembershipAsyncService, IndividualMembershipService
from .manager import ManagerAsyncService, ManagerService
from .manager_membership import ManagerMembershipAsyncService, ManagerMembershipService
from .official import OfficialAsyncService, OfficialService
from .team import TeamAsyncService, TeamService
from .team_membership import TeamMembershipAsyncService, TeamMembershipService

__all__ = [
    "AthleteAsyncService",
    "AthleteService",
    "IndividualMembershipAsyncService",
    "IndividualMembershipService",
    "ManagerAsyncService",
    "ManagerService",
    "ManagerMembershipAsyncService",
    "ManagerMembershipService",
    "OfficialAsyncService",
    "OfficialService",
    "TeamAsyncService",
    "TeamService",
    "TeamMembershipAsyncService",
    "TeamMembershipService",
]



"""Metadata for CLI tool."""

__app_name__ = "mgr"


from leaguemanager.__metadata__ import __version__
from leaguemanager.dependency import LeagueManager

__all__ = ["__app_name__", "__version__", "LeagueManager"]



from datetime import datetime

from sqlalchemy import delete
from sqlalchemy.orm import Session

from leaguemanager.core import get_settings
from leaguemanager.models.base import UUIDBase

settings = get_settings()


def clear_table(session: Session, model: UUIDBase) -> None:
    """Clears table of given model."""
    session.execute(delete(model))


def str_to_iso(date_string: str, format: str):
    """Converts string to datetime object."""
    return datetime.strptime(date_string, format)



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import ParticipationBase, mapper, metadata

if TYPE_CHECKING:
    from leaguemanager.models.baseline import Athlete


@define(slots=False)
class AthleteStats(ParticipationBase):
    """The participation of an individual or team that can win or lose an Fixture.

    Each individual that is participating in an Fixture has a AthleteStats that represents
    an athlete's performance in that given Fixture. This class is very basic/generic as it
    is used to represent a broad set of competitions. (Premium class can provide more sport
    specific stats.)

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        season_id (UUID): ForeignKey to Season.
        phase_id (UUID): ForeignKey to Phase.
        fixture_id (UUID): ForeignKey to Fixture.
        label (str): Description of AthleteStats (i.e., "Mark Smith, Week 4, Explosive FC", etc...)
        points_scored (int): Points scored by the Athlete.
        saves (int): Any "saves" made by a player.
        cautions (str): Any cautions received by player.
        note (str): Miscellaneous info concerning player with this fixture.
        expelled (bool): If player was expelled from the fixture.
        disqualified (bool): If player was disqualified from competing in fixture.
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    season_id: UUID | None = field(default=None)
    phase_id: UUID | None = field(default=None)
    fixture_id: UUID | None = field(default=None)
    label: str | None = field(default=None, validator=validators.max_len(90))
    outcome: str | None = field(
        default=None, validator=validators.optional(validators.in_(("win", "loss", "tie", "undecided")))
    )
    score_for: int | None = field(default=0)
    score_against: int | None = field(default=0)
    alignment: str | None = field(default=None, validator=validators.optional(validators.in_(("home", "away"))))
    standing_points: int | None = field(default=None)


# SQLAlchemy Imperative Mappings

athlete_stats = Table(
    "athlete_stats",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("season_id", SA_UUID, ForeignKey("season.id"), nullable=False),
    Column("phase_id", SA_UUID, ForeignKey("phase.id"), nullable=True),
    Column("fixture_id", SA_UUID, ForeignKey("fixture.id"), nullable=True),
    Column("label", String(90), nullable=False),
    Column("outcome", String(12), nullable=True),
    Column("score_for", Integer, nullable=True),
    Column("score_against", Integer, nullable=True),
    Column("alignment", String(6), nullable=True),
    Column("fixture_standing_points", Integer, nullable=True),
    Column("standing_points", Integer, nullable=True),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    AthleteStats,
    athlete_stats,
    properties={
        "season": relationship("Season", back_populates="athlete_stats"),
        "fixture": relationship("Fixture", back_populates="athlete_stats"),
        "athlete": relationship("Athlete", back_populates="athlete_stats", uselist=False),
    },
)



from .role import RoleAsyncService, RoleSyncService
from .user import UserAsyncService, UserRoleAsyncService, UserRoleSyncService, UserSyncService

__all__ = [
    "RoleAsyncService",
    "RoleSyncService",
    "UserAsyncService",
    "UserRoleAsyncService",
    "UserRoleSyncService",
    "UserSyncService",
]



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import IndividualBase, mapper, metadata


@define(slots=False)
class Official(IndividualBase):
    """An official or referee of competitions/fixtures.

    It is sometimes possible for an Official to fill other IndividualBase roles
    (competing/managing). As a result, an Official is primarily linked to an
    OfficialMembership object. The OfficialMembership object will can  then be used
    to track which League or Leagues they can officiate, and can also track start/end
    dates for said membership.

    Notice that many attributes are inherited from the IndividualBase class.

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        officiating_id (UUID): ForeignKey for Officiating.
        label (str): If present, used to identify the Official.
        license (str): Used to determine if Official has certification and/or license.
        verified (bool): Whether the Official has been verified to officiate.

        first_name (str): Inherited from IndividualBase.
        middle_name (str): Inherited from IndividualBase.
        last_name (str): Inherited from IndividualBase.
        full_name (str): Inherited from IndividualBase.
        alias (str): Inherited from IndividualBase.
        email (str): Inherited from IndividualBase.
        mobile_phone (str): Inherited from IndividualBase.
        birth_date (datetime): Inherited from IndividualBase.
        birth_year (int): Inherited from IndividualBase.
        sex (str): Inherited from IndividualBase.
        gender (str): Inherited from IndividualBase.

        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    officiating_id: UUID | None = field(default=None)
    label: str | None = field(default=None, validator=validators.max_len(40))
    license: str | None = field(default=None, validator=validators.optional(validators.max_len(16)))
    verified: bool = field(default=False)


# SQLAlchemy Imperative Mapping

official = Table(
    "official",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("officiating_id", SA_UUID, ForeignKey("officiating.id"), nullable=True),
    Column("label", String(40), unique=True, nullable=False),
    Column("license", String(16), nullable=False),
    Column("verified", Boolean),
    Column("first_name", String(20), nullable=True),
    Column("middle_name", String(20), nullable=True),
    Column("last_name", String(20), nullable=True),
    Column("full_name", String(64), nullable=True),
    Column("alias", String(40), nullable=True),
    Column("email", String(40), nullable=True),
    Column("mobile_phone", String(16), nullable=True),
    Column("birth_date", DateTime, nullable=True),
    Column("birth_year", Integer, nullable=True),
    Column("sex", String(6), nullable=True),
    Column("gender", String(12), nullable=True),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    Official,
    official,
    properties={
        "officiating": relationship("Officiating", back_populates="officials"),
    },
)



from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field
from sqlalchemy import UUID as _UUID
from sqlalchemy import Column, DateTime, ForeignKey, Table
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import UUIDBase, mapper, metadata

if TYPE_CHECKING:
    from .role import Role
    from .user import User


@define(slots=False)
class UserRole(UUIDBase):
    """A relationship link between a user and a role."""

    user_id: UUID | None = field(default=None)
    role_id: UUID | None = field(default=None)
    assigned_at: datetime = field(factory=lambda: datetime.now(UTC))

    # user_name: AssociationProxy = association_proxy("user", "username")
    # role_name: AssociationProxy = association_proxy("role", "name")


# SQLAlchemy Imperative Mapping

user_role = Table(
    "user_role",
    metadata,
    Column("user_id", _UUID, ForeignKey("user.id"), primary_key=True),
    Column("role_id", _UUID, ForeignKey("role.id"), primary_key=True),
    Column("assigned_at", DateTime, nullable=False),
)

# ORM Relationships

mapper.map_imperatively(
    UserRole,
    user_role,
    properties={
        "user": relationship("User", back_populates="user_roles"),
        "role": relationship("Role", back_populates="user_roles"),
    },
)
